# -*- encoding: utf-8 -*-
# Copyright (c) 2005 Canonical
# Copyright (c) 2004 Conectiva, Inc.
#
# Written by Gustavo Niemeyer <niemeyer@conectiva.com>
#
# This file is part of Smart Package Manager.
#
# Smart Package Manager is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# Smart Package Manager is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Smart Package Manager; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
from smart.backends.rpm.synthesis import URPMISynthesisLoader
from smart.backends.rpm.header import URPMILoader
from smart.util.filetools import getFileDigest, compareFiles
from smart.const import SUCCEEDED, FAILED, ALWAYS, NEVER
from smart.channel import PackageChannel
from smart import *
import posixpath
import re
import os
import string
from time import time

class URPMIChannel(PackageChannel):
    # It's important for the default to be here so that old pickled
    # instances which don't have these attributes still work fine.
    _mirrorurl = None

    def __init__(self, baseurl, mirrorurl, hdlurl, infourl, *args):
        super(URPMIChannel, self).__init__(*args)

        self._baseurl = baseurl
        self._mirrorurl = mirrorurl
        self._hdlurl = hdlurl
        self._infourl = infourl

        if not self._mirrorurl:
            self._setURLs()


    def _setURLs(self):
        if self._hdlurl:
            if self._hdlurl[0] != "/" and ":/" not in self._hdlurl:
                self._hdlurl = posixpath.join(self._baseurl, self._hdlurl)
        else:
            self._hdlurl = posixpath.join(self._baseurl, "hdlist.cz")
        if self._infourl:
            if self._infourl[0] != "/" and ":/" not in self._infourl:
                self._infourl = posixpath.join(self._baseurl, self._infourl)
        else:
            self._infourl = posixpath.join(self._baseurl, "info.xml.lzma")

        self._compareurl = self._hdlurl


    def getCacheCompareURLs(self):
        return [self._compareurl]

    def getFetchSteps(self):
        return 3
    
    def getMirrorListURL(self):
        if self._mirrorurl == "$MIRRORLIST":
            fp = open("/etc/product.id")
            items = fp.readline().strip().split(",")
            fp.close()
            productid = {}
            for item in items:
                key,entry = item.split("=")
                productid[key] = entry.lower()
            self._mirrorurl = "http://api.mandriva.com/mirrors/%s.%s.%s.list" % (productid["type"], productid["version"], productid["arch"])
        elif self._mirrorurl.find("$RELEASE") != -1 or self._mirrorurl.find("$ARCH") != -1:
            fp = open("/etc/release")
            line = fp.readline().strip().lower()
            fp.close()
            release = line.split(")")[0].split("(")[1]
            arch = line.split("for ")[1] 
            self._mirrorurl = self._mirrorurl.replace("$RELEASE", release).replace("$ARCH", arch)
        return self._mirrorurl

    def getMirrors(self, fetcher, progress):
        if not self._mirrorurl:
            return None

        dir = fetcher.getLocalDir()
        prefix = fetcher.getLocalPathPrefix()
        fetcher.setLocalPathPrefix("")
        mirrordir = posixpath.join(os.path.realpath(dir + os.pardir), "mirrorlists")
        if not posixpath.exists(mirrordir):
                os.mkdir(mirrordir)
        fetcher.setLocalDir(mirrordir, mangle=True)

        item = fetcher.enqueue(self.getMirrorListURL())
        path = fetcher.getLocalPath(item)
        if not posixpath.exists(path) \
                or time() > (posixpath.getmtime(path) + 24 * 60 * 60):
            fetcher.run(progress=progress)
            failed = item.getFailedReason()
        else:
            failed = False

        fetcher.setLocalPathPrefix(prefix)
        fetcher.setLocalDir(dir, mangle=False)
        fetcher.reset()

        if failed:
            return None

        from smart.util.geolocate import GeoLocate
        geoloc = GeoLocate()
        fp = open(path)
        mirrors = []
        for line in fp.readlines():
            mirror = {"country": None, "continent": None}
            for item in line.split(","):
                key, entry = item.split("=")
                mirror[key] = entry.strip()
            scheme = mirror["url"].split(":")[0]
            if not fetcher.getHandler(scheme, None) or mirror["type"] != "distrib":
                continue
            mirror["proximity"] = geoloc.getProximity(float(mirror["latitude"]), float(mirror["longitude"]), randomize=True, country=mirror["country"], continent=mirror["continent"])
            mirrors.append(mirror)
        fp.close()
        mirrors.sort(cmp=lambda x,y: cmp(x["proximity"], y["proximity"]))

        return mirrors

    def fetch(self, fetcher, progress):

        fetcher.reset()

        mirrors = self.getMirrors(fetcher, progress)
        if mirrors:
            for mirror in mirrors:
                if mirror == mirrors[0]:
                    self._baseurl = mirror["url"]
                else:
                    sysconf.add(("mirrors", self._baseurl), mirror["url"], unique=True)

            self._setURLs()
            fetcher.enqueue(posixpath.join(self._baseurl, "media/media_info/media.cfg"))
            fetcher.run()
            fetcher.reset()
        if self.getAlias().startswith("urpmisync-") and sysconf.get("sync-urpmi-medialist"):
            urpmiDir = sysconf.get("sync-urpmi-mediadir")
            urpmiMediaDir = posixpath.join(urpmiDir, self.getName())

            # We'll force migration to new layout of urpmi media rather than
            # implementing support for old layout
            if not posixpath.exists(urpmiMediaDir):
                os.mkdir(urpmiMediaDir)
            # Here we'll also force moving of existing data to new layout and
            # also to get rid of any old leftovers..
            for filename in ["changelog.%s.xml.lzma", "files.%s.xml.lzma", \
                    "info.%s.xml.lzma", "MD5SUM.%s", "hdlist.%s.cz", \
                    "synthesis.hdlist.%s.cz"]:
                filename %= self.getName()
                oldfile = posixpath.join(urpmiDir, filename)
                if not os.path.exists(oldfile):
                    continue
                tmp = filename.split("."+self.getName())
                newfile = posixpath.join(urpmiMediaDir, string.join(tmp, ""))
                if compareFiles(oldfile, newfile) or \
                        os.path.getmtime(oldfile) < os.path.getmtime(oldfile) or\
                        os.path.getsize == 0:
                    os.unlink(oldfile)
                else:
                    os.rename(oldfile, newfile)

            fetcher.setLocalDir(urpmiMediaDir)
            fetcher.setLocalPathPrefix("")

        self._compareurl = self._hdlurl

        hdlbaseurl, basename = os.path.split(self._hdlurl)

        md5url = posixpath.join(hdlbaseurl, "MD5SUM")
        item = fetcher.enqueue(md5url)
        fetcher.run(progress=progress)

        hdlmd5 = None
        failed = item.getFailedReason()
        if not failed:
            self._compareurl = md5url
            digest = getFileDigest(item.getTargetPath())

            if digest == self._digest:
                progress.add(2)
                return True

            basename = posixpath.basename(self._hdlurl)
            for line in open(item.getTargetPath()):
                line = line.strip()
                if line:
                    md5, name = line.split()
                    if name == basename:
                        hdlmd5 = md5
                        break

        fetcher.reset()
        hdlitem = fetcher.enqueue(self._hdlurl, md5=hdlmd5, uncomp=False)

        fetcher.run(progress=progress)

        if hdlitem.getStatus() == FAILED:
            hdfailed = hdlitem.getFailedReason()
            if fetcher.getCaching() is NEVER:
                # Try reading reconfig.urpmi (should give new path)
                fetcher.reset()
                reconfigurl = posixpath.join(hdlbaseurl, "reconfig.urpmi")
                reconfigitem = fetcher.enqueue(reconfigurl)
                fetcher.run(progress=progress)
                if reconfigitem.getStatus() == FAILED:
                    refailed = reconfigitem.getFailedReason()
                    if fetcher.getCaching() is NEVER:
                        lines = [_("Failed acquiring information for '%s':") % self,
                            u"%s: %s" % (hdlitem.getURL(), hdfailed),
                            u"%s: %s" % (reconfigitem.getURL(), refailed)]
                        raise Error, "\n".join(lines)
                    return False
                else:
                    # Need to inject "/" at the end to avoid buggy urls
                    if not hdlbaseurl.endswith("/"): hdlbaseurl += "/"
                    for line in open(reconfigitem.getTargetPath()):
                        if line.startswith("#"): pass
                        elif line:
                            splitline = line.split()
                            arch = os.uname()[4]
                            if arch == "i686": arch = "i586"
                            reconfarch = re.sub("\$ARCH", arch, splitline[1])
                            reconfpath = re.sub(splitline[0] + "$", reconfarch, hdlbaseurl)
                            sysconf.set(("channels", self.getAlias(), \
                                        "baseurl"), reconfpath)
                            self._hdlurl = os.path.join(reconfpath, basename)
                    return self.fetch(fetcher, progress)
            return False
        else:
            localpath = hdlitem.getTargetPath()
            digestpath = None
            
            # Packdrakeng archive has a 64 byte trailer at end, if present
            # we're using hdlist.cz
            localhdl = open(localpath)
            localhdl.seek(0,2)
            if localhdl.tell() > 64:
                localhdl.seek(-64,2)
            trailer = localhdl.read()
            localhdl.close()

            if self._infourl and trailer[:3] + trailer[-3:] != "cz[]cz":
                infoitem = fetcher.enqueue(self._infourl, uncomp=False)
                fetcher.run(progress=progress)
            else:
                infoitem = None

            if infoitem and infoitem.getStatus() == SUCCEEDED:
                if self._compareurl == self._hdlurl:
                    self._compareurl = infourl
                    digestpath = localpath
                infopath = infoitem.getTargetPath()
            else:
                infopath = None
                if self._compareurl == self._hdlurl:
                    digestpath = localpath
            if digestpath:
                digest = getFileDigest(digestpath)
                if digest == self._digest:
                    return True
            self.removeLoaders()

            if len(trailer) >= 64 and trailer[:3] + trailer[-3:] == "cz[]cz":
                uncompressor = fetcher.getUncompressor()
                uncomphandler = uncompressor.getHandler(localpath)
                uncomphandler.uncompress(localpath)
                localpath = localpath[:-3]
                loader = URPMILoader(localpath, self._baseurl, None)
            else:
                loader = URPMISynthesisLoader(localpath, self._baseurl, infopath)
                                
            loader.setChannel(self)
            self._loaders.append(loader)

        self._digest = digest

        return True

def create(alias, data):
    return URPMIChannel(data["baseurl"],
                        data["mirrorurl"],
                        data["hdlurl"],
                        data["infourl"],
                        data["type"],
                        alias,
                        data["name"],
                        data["manual"],
                        data["removable"],
                        data["priority"])

# vim:ts=4:sw=4:et
